# -*- coding: utf-8 -*-
import logging
import coloredlogs


def init_log(logger, loglevel='WARNING'):
    """Initializes the log file in the proper format.
    Arguments:
        filename (str): Path to a file. Or None if logging is to
                         be disabled.
        loglevel (str): Determines the level of the log output.
    """
    if loglevel:
        logger.setLevel(getattr(logging, loglevel))
        coloredlogs.install(level=loglevel)

    # be more strict about logging from requests package ;)
    logging.getLogger('requests').setLevel(logging.WARNING)



# -*- coding: utf-8 -*-
import datetime
import json
import re

from six import StringIO

from cglims.constants import READS_PER_1X


def json_serial(obj):
    """JSON serializer for objects not serializable by default json code"""
    if isinstance(obj, datetime.datetime):
        serial = obj.isoformat()
        return serial
    raise TypeError('Type not serializable')


def jsonify(data, pretty=False):
    """Serialize Model to JSON."""
    kwargs = dict(indent=4, sort_keys=True) if pretty else dict()
    return json.dumps(data, default=json_serial, **kwargs)


def fix_dump(dump, indentSize=2):
    stream = StringIO(dump)
    out = StringIO()
    pattern = re.compile('(\s*)([^:]*)(:*)')
    last = None

    prefix = 0
    for line in stream:
        indent, key, colon = pattern.match(line).groups()
        if indent == "" and key[0] != '-':
            prefix = 0
        if last:
            if len(last[0]) == len(indent) and last[2] == ':':
                if all([not last[1].startswith('-'),
                        line.strip().startswith('-')]):
                    prefix += indentSize
        out.write(" " * prefix + line)
        last = indent, key, colon
    return out.getvalue()


def ordered_reads(app_tag):
    """Calculate ordered number of reads."""
    type_id = app_tag[-4]
    number = int(app_tag[-3:])
    if type_id == 'R':
        return number * 1000000
    elif type_id == 'K':
        return number * 1000
    elif type_id == 'C':
        # expect WGS
        return number * READS_PER_1X
    else:
        raise ValueError("unknown read type id: {}".format(type_id))



# -*- coding: utf-8 -*-
"""
DEPRECATED.
"""
import logging
import io
import copy
import csv

from .config import get_capture_kit, CAPTUREKIT_MAP
from .exc import MissingLimsDataException
from .apptag import ApplicationTag


PHENOTYPE_MAP = dict(affected='2', unaffected='1', unknown='0')
SEX_MAP = dict(M='1', F='2', Unknown='other', unknown='other')
MANDATORY_HEADERS = ['Family ID', 'Individual ID', 'Paternal ID',
                     'Maternal ID', 'Sex', 'Phenotype']
EXTRA_HEADERS = ['Clinical_db', 'Capture_kit', 'display_name',
                 'Sequencing_type']

log = logging.getLogger(__name__)


def get_sampleid(lims_sample, key='Clinical Genomics ID'):
    """Get the expected LIMS or Clinical Genomics ID."""
    return lims_sample.udf.get(key) or lims_sample.id


def old_capture_kit(lims, lims_sample, udf_key='Capture Library version',
                    udf_kitkey='Capture Library version'):
    """Figure out which capture kit has been used for the sample."""
    hybrizelib_id = '33'
    if udf_key in dict(lims_sample.udf.items()):
        log.debug('prefer capture kit annotated on the sample level')
        capture_kit = lims_sample.udf[udf_key]
    else:
        artifacts = lims.get_artifacts(samplelimsid=lims_sample.id,
                                       type='Analyte')
        capture_kit = None
        for artifact in artifacts:
            if artifact.parent_process:
                if artifact.parent_process.type.id == hybrizelib_id:
                    try:
                        capture_kit = artifact.parent_process.udf[udf_kitkey]
                    except KeyError:
                        log.warn('capture kit not found on expected process')
                        continue
                    break

        if capture_kit is None:
            raise MissingLimsDataException("No capture kit annotated: {}"
                                           .format(lims_sample.id))

    return CAPTUREKIT_MAP[capture_kit.strip()]


def make_pedigree(api, lims_samples, family_id=None, gene_panel=None,
                  internalize=True):
    """String together all individual steps to create a pedigree."""
    # extract information about samples
    samples = [convert_sample(api, lims_sample, gene_panel=gene_panel)
               for lims_sample in lims_samples
               if lims_sample.udf.get('cancelled') != 'yes']
    if internalize:
        samples = internalize_ids(samples)
    if family_id:
        raw_samples = samples
        samples = []
        for sample in raw_samples:
            sample['Family ID'] = family_id
            samples.append(sample)
    # serialize to CSV PED file
    rows = serialize(samples)
    return "#{}".format(rows)


def convert_sample(api, lims_sample, gene_panel=None):
    """Extract information from LIMS samples that relate to pedigree."""
    app_tag = ApplicationTag(lims_sample.udf['Sequencing Analysis'])
    affection_status = lims_sample.udf['Status'].lower()
    sex_letter = lims_sample.udf['Gender']
    if gene_panel:
        gene_panels = gene_panel
    else:
        try:
            gene_panels = lims_sample.udf['Gene List']
            if ':' in gene_panels:
                log.warn("wrong separator in 'Gene List': %s", gene_panels)
                udf_key = 'Gene List'
                new_value = gene_panels.replace(':', ';')
                lims_sample.udf[udf_key] = new_value
                log.info("updating %s: '%s' -> '%s'", lims_sample.id,
                         gene_panels, new_value)
                lims_sample.put()
                gene_panels = new_value
        except KeyError:
            message = "{}: 'Gene List'".format(lims_sample.id)
            raise MissingLimsDataException(message)
    try:
        ped_phenotype = PHENOTYPE_MAP[affection_status]
    except KeyError:
        msg = "{}: 'Status' -> '{}'".format(lims_sample.id, affection_status)
        raise MissingLimsDataException(msg)
    data = {
        'Family ID': lims_sample.udf['familyID'],
        'Individual ID': lims_sample.name,
        'Paternal ID': lims_sample.udf.get('fatherID', '0'),
        'Maternal ID': lims_sample.udf.get('motherID', '0'),
        'Sex': SEX_MAP[sex_letter],
        'Phenotype': ped_phenotype,
        'Clinical_db': gene_panels,
        'Sequencing_type': app_tag.sequencing_type,
        'internal_id': get_sampleid(lims_sample),
    }

    # fetch capture kit if sample is exome sequenced
    if data['Sequencing_type'] == 'wes':
        data['Capture_kit'] = get_capture_kit(api, lims_sample)
    else:
        data['Capture_kit'] = None

    return data


def internalize_ids(samples):
    """Replace customer sample ids with internal ids."""
    sample_map = {sample['Individual ID']: sample for sample in samples}
    for external_id, sample in sample_map.items():
        new_sample = copy.deepcopy(sample)
        new_sample['Individual ID'] = sample['internal_id']
        new_sample['display_name'] = external_id

        parent_fields = ['Paternal ID', 'Maternal ID']
        for parent_field in parent_fields:
            parent_id = new_sample[parent_field]
            if parent_id != '0':
                new_sample[parent_field] = sample_map[parent_id]['internal_id']
        yield new_sample


def serialize(sample_dicts, headers=None, extra=None):
    """Serialize a list of samples into PED output to a file.

    Args:
        sample_dicts (List[dict]): list of dicts with values
        headers (Optional[list]): order of headers to output
    """
    output = io.BytesIO()
    headers = headers or MANDATORY_HEADERS
    extra = extra or EXTRA_HEADERS
    all_headers = headers + extra
    dict_writer = csv.DictWriter(output, all_headers, extrasaction='ignore',
                                 dialect='excel-tab', lineterminator='\n')
    dict_writer.writeheader()
    dict_writer.writerows(sample_dicts)
    return output.getvalue()



# -*- coding: utf-8 -*-

import warnings

from cglims.constants import READS_PER_1X
from cglims.exc import UnknownSequencingTypeError

PANELS = set(['EXO', 'EXT', 'MHP', 'EFT', 'CCP', 'EXX', 'EXL'])
TARGETED = set(['MHP', 'EFT', 'CCP'])
WHOLEGENOME = set(['WGS', 'WGT', 'WGL', 'MWG', 'MWL', 'MWX', 'MET', 'MEL', 'WGX', 'MWR', 'VWG', 'MWM', 'VWM'])
ANALYSIS_ONLY = set(['EXX', 'WGX'])
MICROBIAL = set(['MWX', 'MWG', 'MWL', 'MWR', 'VWG', 'MWM', 'VWM'])
RNA = set(['RNA', 'RNL'])
HUMAN = (PANELS | WHOLEGENOME | ANALYSIS_ONLY) - MICROBIAL - RNA


class ApplicationTag(str):

    def __init__(self, raw_tag):
        super(ApplicationTag, self).__init__()
        self = raw_tag

    @property
    def application(self):
        """Get the application part of the tag."""
        return self[:3]

    @property
    def sequencing(self):
        """DEPRICATED: replaced by application()."""
        warnings.warn('Deprecated: use application() instead', DeprecationWarning, stacklevel=2)
        return self[:3]

    @property
    def library_prep(self):
        """Get the library preparation part of the tag."""
        return self[3:6]

    @property
    def is_human(self):
        """Determine if human sequencing."""
        return self.application in HUMAN

    @property
    def is_panel(self):
        """Determine if sequencing if sequence capture."""
        return self.application in PANELS

    @property
    def analysis_type(self):
        """Return analysis type from tag."""
        warnings.warn('Deprecated: use sequencing_type() instead', DeprecationWarning, stacklevel=2)
        return self.sequencing_type

    @property
    def category(self):
        """Return the category of application."""
        category = self.sequencing_type
        if self.is_external:
            category = "{}-ext".format(category)
        return category

    @property
    def is_microbial(self):
        """Determine if the order is for regular microbial samples."""
        return self.application in MICROBIAL

    @property
    def is_rna(self):
        """Determine if the order is for RNAseq samples."""
        return self.application in RNA

    @property
    def sequencing_type(self):
        """parse application type to figure out type of sequencing."""
        if self.application in WHOLEGENOME:
            return 'wgs'
        elif self.sequencing in TARGETED:
            return 'tga'
        elif self.application in PANELS:
            return 'wes'
        else:
            raise UnknownSequencingTypeError("Application '{}' is unknown.".format(self.application))

    @property
    def sequencing_type_mip(self):
        """Convert sequencing type to MIP specific version."""
        if self.sequencing_type == 'tga':
            return 'wes'
        else:
            return self.sequencing_type

    @property
    def is_external(self):
        """ Determines whether or not a sample is externally sequenced.

        Returns (bool): True when external, False otherwise
        """

        if self.application.endswith('X'):
            return True
        return False

    @property
    def reads(self):
        """Calculate ordered number of reads."""
        type_id = self[-4]
        number = int(self[-3:])
        if type_id == 'R':
            return number * 1000000
        elif type_id == 'K':
            return number * 1000
        elif type_id == 'C':
            if self.is_panel:
                raise ValueError("can't convert coverage for panels")
            return number * 10000000 # but should be: number * READS_PER_1X
        else:
            raise ValueError("unknown read type id: {}".format(type_id))



# -*- coding: utf-8 -*-
import logging

import click
import yaml

from cglims import api
from cglims.api import ClinicalSample
from cglims.apptag import UnknownSequencingTypeError
from cglims.config import make_config, CAPTUREKIT_MAP, relevant_samples
from cglims.pedigree import make_pedigree
from cglims.panels import convert_panels
from .utils import jsonify, fix_dump

CAPTUREKITS = CAPTUREKIT_MAP.values()
log = logging.getLogger(__name__)


@click.command()
@click.option('-g', '--gene-panel', help='custom gene panel')
@click.option('-f', '--family-id', help='custom family id')
@click.option('-s', '--samples', multiple=True, help='included samples')
@click.argument('customer_family', nargs=2, required=False)
@click.pass_context
def pedigree(context, gene_panel, family_id, samples, customer_family):
    """DEPRECATED: Create pedigree from LIMS."""
    lims_api = api.connect(context.obj)
    if customer_family:
        lims_samples = lims_api.case(*customer_family)
    elif samples:
        lims_samples = [lims_api.sample(sample_id) for sample_id in samples]
    else:
        click.echo("you need to provide customer+family or samples")
        context.abort()
    content = make_pedigree(lims_api, lims_samples, family_id=family_id,
                            gene_panel=gene_panel)
    click.echo(content)


@click.command()
@click.option('-g', '--gene-panel', help='custom gene panel')
@click.option('-f', '--family-id', help='custom family id')
@click.option('-s', '--samples', multiple=True, help='included samples')
@click.option('-c', '--capture-kit', type=click.Choice(CAPTUREKITS),
              help='custom capture kit')
@click.option('--force', is_flag=True, help='skip sanity checks')
@click.argument('raw_case_id')
@click.pass_context
def config(context, gene_panel, family_id, samples, capture_kit, force,
           raw_case_id):
    """Create pedigree YAML file from LIMS data."""
    if '--' in raw_case_id:
        case_id, ext = raw_case_id.split('--', 1)
    else:
        case_id, ext = raw_case_id, None
    customer, family = case_id.split('-', 1)

    lims_api = api.connect(context.obj)
    gene_panels = [gene_panel] if gene_panel else None

    if samples:
        lims_samples = [lims_api.sample(sample_id) for sample_id in samples]
    else:
        lims_samples = lims_api.case(customer, family)

    included_samples = relevant_samples(lims_samples)
    data = make_config(lims_api, included_samples, family_id=family_id,
                       gene_panels=gene_panels, capture_kit=capture_kit, force=force)

    if ext:
        # handle cases with e.g. downsampled data
        data['family'] = '--'.join([data['family'], ext])
        for sample in data['samples']:
            sample['sample_id'] = '--'.join([sample['sample_id'], ext])

    # handle single sample cases with 'unknown' phenotype
    if len(data['samples']) == 1:
        if data['samples'][0]['phenotype'] == 'unknown':
            log.info("setting 'unknown' phenotype to 'unaffected'")
            data['samples'][0]['phenotype'] = 'unaffected'

    dump = yaml.safe_dump(data, default_flow_style=False, allow_unicode=True)
    click.echo(fix_dump(dump))


@click.command()
@click.option('-c', '--condense', is_flag=True, help='condense output')
@click.option('-p', '--project', is_flag=True, help='identifier is a project')
@click.option('-n', '--external', is_flag=True, help='identifier is the customer sample name')
@click.option('-m', '--minimal', is_flag=True, help='output minimal information')
@click.option('--all', '--all-samples', is_flag=True,
              help='include cancelled/tumor samples')
@click.argument('raw_identifier')
@click.argument('field', required=False)
@click.pass_context
def get(context, condense, project, external, minimal, raw_identifier, field, all_samples):
    """Get information from LIMS: either sample or family samples."""
    if '--' in raw_identifier:
        identifier, ext = raw_identifier.split('--', 1)
    else:
        identifier, ext = raw_identifier, None

    lims = api.connect(context.obj)
    if project:
        lims_samples = lims.get_samples(projectlimsid=identifier)
    elif identifier.startswith('cust'):
        # look up samples in a case
        lims_samples = lims.case(*identifier.split('-', 1))
    elif external:
        lims_samples = lims.get_samples(name=identifier)
    else:
        # look up a single sample
        is_cgid = True if identifier[0].isdigit() else False
        lims_samples = [lims.sample(identifier, is_cgid=is_cgid)]

    if len(lims_samples) > 1 and not all_samples:
        # filter out tumor and cancelled samples
        lims_samples = relevant_samples(lims_samples)

    for lims_sample in lims_samples:
        sample_obj = ClinicalSample(lims_sample)
        data = sample_obj.to_dict(minimal=minimal)
        data['sample_id'] = "{}--{}".format(data['sample_id'], ext) if ext else data['sample_id']
        data['case_id'] = "{}--{}".format(data['case_id'], ext) if ext else data['case_id']

        if field:
            if field not in data:
                log.error("can't find UDF on sample: %s", field)
                context.abort()
            elif isinstance(data[field], list):
                for item in data[field]:
                    click.echo(item)
            else:
                click.echo(data[field])
        else:
            if condense:
                dump = jsonify(data)
            else:
                raw_dump = yaml.safe_dump(data, default_flow_style=False,
                                          allow_unicode=True)
                dump = fix_dump(raw_dump)
                click.echo(click.style('>>> Sample: ', fg='red'), nl=False)
                click.echo(click.style(data['id'], bold=True, fg='red'))
                if data.get('cancelled') == 'yes':
                    click.echo(click.style('CANCELLED', bold=True, fg='yellow'))
            click.echo(dump)


@click.command()
@click.argument('lims_id')
@click.argument('field_key')
@click.argument('new_value')
@click.pass_context
def update(context, lims_id, field_key, new_value):
    """Update a UDF for a sample."""
    lims = api.connect(context.obj)
    lims_sample = lims.sample(lims_id)
    old_value = lims_sample.udf.get(field_key, 'N/A').encode('utf-8')
    click.echo("about to update sample: {}".format(lims_sample.id))
    message_tmlt = "are you sure you want to change '{}': '{}' -> '{}'"
    if click.confirm(message_tmlt.format(field_key, old_value, new_value)):
        lims_sample.udf[field_key] = new_value
        lims_sample.put()


@click.command()
@click.argument('sample_id')
@click.pass_context
def fillin(context, sample_id):
    """Fill in defaults for a LIMS sample."""
    lims_api = api.connect(context.obj)
    lims_sample = lims_api.sample(sample_id)
    click.echo("filling in defaults...")
    set_defaults(lims_sample)
    lims_sample.put()
    click.echo("saved new defaults")


def set_defaults(lims_sample):
    """Set default values for required UDFs."""
    log.info("setting defaults for required fields")
    lims_sample.udf['Concentration (nM)'] = 'NA'
    lims_sample.udf['Volume (uL)'] = 'NA'
    lims_sample.udf['Capture Library version'] = 'NA'
    lims_sample.udf['Strain'] = 'NA'
    lims_sample.udf['Source'] = 'other'
    lims_sample.udf['Index type'] = 'NA'
    lims_sample.udf['Index number'] = 'NA'
    lims_sample.udf['Sample Buffer'] = 'NA'
    lims_sample.udf['Reference Genome Microbial'] = 'NA'
    gender = lims_sample.udf.get('Gender')
    if gender:
        lims_sample.udf['Gender'] = gender.upper()
    else:
        lims_sample.udf['Gender'] = 'unknown'
    if lims_sample.udf.get('Status'):
        lims_sample.udf['Status'] = lims_sample.udf.get('Status').lower()

    if lims_sample.udf.get('Source') == 'Blod':
        log.info("updating 'Source': 'Blod' => 'blood'")
        lims_sample.udf['Source'] = 'blood'

    if 'priority' in lims_sample.udf:
        if lims_sample.udf.get('priority') == 'prioriterad':
            log.info("updating 'priority': 'prioriterad' => 'priority'")
            lims_sample.udf['priority'] = 'priority'
        else:
            lims_sample.udf['priority'] = lims_sample.udf['priority'].lower()
    else:
        log.info("missing 'priority' => setting to 'standard'")
        lims_sample.udf['priority'] = 'standard'

    process_only = lims_sample.udf.get('Process only if QC OK')
    if process_only == 'Ja':
        log.info("translating 'QC OK' field: 'Ja' => 'yes'")
        lims_sample.udf['Process only if QC OK'] = 'yes'
    elif process_only == 'Nej':
        log.info("translating 'QC OK' field: 'Nej' => 'no'")
        lims_sample.udf['Process only if QC OK'] = 'no'
    elif process_only is None:
        log.info("setting 'QC OK' field to default: 'NA'")
        lims_sample.udf['Process only if QC OK'] = 'NA'


@click.command()
@click.argument('customer')
@click.argument('default_panels', nargs=-1)
@click.pass_context
def panels(context, customer, default_panels):
    """Convert between default panels and gene list panels."""
    for panel_id in convert_panels(customer, default_panels):
        click.echo(panel_id)


@click.command()
@click.option('-d', '--delivered', is_flag=True, help='check if sample is delivered')
@click.argument('lims_id')
@click.pass_context
def sample(context, delivered, lims_id):
    """Fetch information about a sample."""
    lims_api = api.connect(context.obj)
    if delivered:
        delivery_date = lims_api.is_delivered(lims_id)
        if delivery_date:
            click.echo(delivery_date)
        else:
            log.error("sample not yet delivered")
            context.abort()



# -*- coding: utf-8 -*-
import logging

from .exc import MissingLimsDataException
from .apptag import ApplicationTag
from .panels import convert_panels

SEX_MAP = dict(M='male', F='female', Unknown='unknown', unknown='unknown')
CAPTUREKIT_MAP = {'Agilent Sureselect CRE': 'Agilent_SureSelectCRE.V1',
                  'SureSelect CRE': 'Agilent_SureSelectCRE.V1',
                  'Agilent Sureselect V5': 'Agilent_SureSelect.V5',
                  'SureSelect Focused Exome': 'Agilent_SureSelectFocusedExome.V1',
                  'other': 'Agilent_SureSelectCRE.V1'}
LATEST_CAPTUREKIT = 'Agilent_SureSelectCRE.V1'

log = logging.getLogger(__name__)


def relevant_samples(lims_samples):
    """Filter out relevant samples for analysis."""
    included = (lims_sample for lims_sample in lims_samples
                if (lims_sample.udf.get('cancelled') != 'yes' and
                    lims_sample.udf.get('tumor') != 'yes' and
                    lims_sample.udf.get('exclude analysis') != 'yes'))
    return included


def basic_config(lims_api, customer_id, family_id):
    """Generate data for a basic config."""
    lims_samples = lims_api.case(customer_id, family_id)
    included_samples = relevant_samples(lims_samples)
    data = make_config(lims_api, included_samples, family_id=family_id)
    # handle single sample cases with 'unknown' phenotype
    if len(data['samples']) == 1:
        if data['samples'][0]['phenotype'] == 'unknown':
            log.info("setting 'unknown' phenotype to 'unaffected'")
            data['samples'][0]['phenotype'] = 'unaffected'
    return data


def gather_data(lims_sample):
    """Gather analysis/pedigree data about a sample."""
    customer = lims_sample.udf['customer']
    family_id = lims_sample.udf['familyID']
    sex_letter = lims_sample.udf['Gender']
    app_tag = ApplicationTag(lims_sample.udf['Sequencing Analysis'])

    data = {
        'sample_id': get_sampleid(lims_sample),
        'sample_name': lims_sample.name,
        'sex': SEX_MAP[sex_letter],
        'phenotype': lims_sample.udf['Status'].lower(),
        'analysis_type': app_tag.sequencing_type_mip,
        'expected_coverage': expected_coverage(app_tag),
    }
    for parent_field in ('fatherID', 'motherID'):
        parent_id = lims_sample.udf.get(parent_field)
        if parent_id and parent_id != '0':
            data[parent_field.replace('ID', '')] = parent_id
    return customer, family_id, data


def expected_coverage(app_tag):
    """Parse out the expected coverage from the app tag."""
    read_part = app_tag[-4:]
    if read_part.startswith('C'):
        return int(read_part[1:]) * 0.87
    elif read_part.startswith('R'):
        # target reads expressed in millions
        return int(read_part[1:]) * 1.5 * 0.87
    else:
        raise ValueError("unexpected app tag: %s", app_tag)


def make_config(lims_api, lims_samples, customer=None, family_id=None,
                gene_panels=None, capture_kit=None, force=False):
    """Make the config for all samples."""
    samples_data = []
    customers = set()
    families = set()
    all_panels = set()
    for lims_sample in lims_samples:
        sample_customer, sample_family, data = gather_data(lims_sample)
        customers.add(sample_customer)
        families.add(sample_family)

        # fetch capture kit if sample is exome sequenced
        if data['analysis_type'] == 'wes':
            data['capture_kit'] = capture_kit or get_capture_kit(lims_api, lims_sample)
        else:
            # wgs: just fill in a default capture kit for coverage analysis
            data['capture_kit'] = LATEST_CAPTUREKIT

        sample_panels = get_genepanels(lims_sample)
        for sample_panel in sample_panels:
            all_panels.add(sample_panel)

        # add mandatory columns
        data['father'] = data.get('father', 0)
        data['mother'] = data.get('mother', 0)

        samples_data.append(data)

    samples_data = list(internalize_ids(samples_data))
    if not force:
        check_relations(samples_data)

    if customer is None:
        assert len(customers) == 1, "conflicting custs: {}".format(customers)
        customer = customers.pop()
    if family_id is None:
        assert len(families) == 1, "conflicting families: {}".format(families)
        family_id = families.pop()

    gene_panels = gene_panels or list(all_panels)
    case_data = {
        'owner': customer,
        'family': family_id,
        'default_gene_panels': gene_panels,
        'gene_panels': convert_panels(customer, gene_panels),
        'samples': samples_data,
    }
    return case_data


def get_genepanels(lims_sample):
    try:
        genepanel_str = lims_sample.udf['Gene List']
        if ':' in genepanel_str:
            log.warn("wrong separator in 'Gene List': %s", genepanel_str)
            udf_key = 'Gene List'
            new_value = genepanel_str.replace(':', ';')
            lims_sample.udf[udf_key] = new_value
            log.info("updating %s: '%s' -> '%s'", lims_sample.id,
                     genepanel_str, new_value)
            lims_sample.put()
            genepanel_str = new_value
    except KeyError:
        message = "{}: 'Gene List'".format(lims_sample.id)
        raise MissingLimsDataException(message)

    gene_panels = set(gene_panel.strip() for gene_panel in genepanel_str.split(';'))
    additional = lims_sample.udf.get('Additional Gene List')
    if additional:
        gene_panels.add(additional.strip())

    return list(gene_panels)


def get_sampleid(lims_sample, key='Clinical Genomics ID'):
    """Get the expected LIMS or Clinical Genomics ID."""
    return lims_sample.udf.get(key) or lims_sample.id


def get_capture_kit(lims, lims_sample, udf_key='Capture Library version',
                    udf_kitkey='SureSelect capture library/libraries used'):
    """Figure out which capture kit has been used for the sample."""
    hybrizelib_id = '669'
    udfs = dict(lims_sample.udf.items())
    sample_capture_kit = udfs.get(udf_key)
    if sample_capture_kit and sample_capture_kit != 'NA':
        log.debug('prefer capture kit annotated on the sample level')
        capture_kit = lims_sample.udf[udf_key]
    else:
        artifacts = lims.get_artifacts(samplelimsid=lims_sample.id,
                                       type='Analyte')
        capture_kit = None
        for artifact in artifacts:
            if artifact.parent_process:
                if artifact.parent_process.type.id == hybrizelib_id:
                    try:
                        capture_kit = artifact.parent_process.udf[udf_kitkey]
                    except KeyError:
                        log.warn('capture kit not found on expected process')
                        continue
                    break

        if capture_kit is None:
            raise MissingLimsDataException("No capture kit annotated: {}"
                                           .format(lims_sample.id))

    return CAPTUREKIT_MAP[capture_kit.strip()]


def internalize_ids(samples):
    """Replace customer sample ids with internal ids."""
    sample_map = {sample['sample_name']: sample for sample in samples}
    for external_id, sample_data in sample_map.items():
        parent_fields = ['father', 'mother']
        for parent_field in parent_fields:
            parent_id = sample_data.get(parent_field)
            if parent_id and parent_id != '0':
                sample_data[parent_field] = sample_map[parent_id]['sample_id']
        yield sample_data


def check_relations(samples):
    """Check parental relations between samples."""
    sexes = {sample['sample_id']: sample['sex'] for sample in samples}
    for sample in samples:
        if sample['father'] != 0:
            if sample['father'] not in sexes:
                raise ValueError("father not referenced: %s", sample['father'])
            elif sexes[sample['father']] != 'male':
                sex = sexes[sample['father']]
                raise ValueError("father sex incorrect: %s, %s", sample['father'], sex)

        if sample['mother'] != 0:
            if sample['mother'] not in sexes:
                raise ValueError("mother not referenced: %s", sample['mother'])
            elif sexes[sample['mother']] != 'female':
                sex = sexes[sample['mother']]
                raise ValueError("mother sex incorrect: %s, %s", sample['mother'], sex)



# -*- coding: utf-8 -*-
import logging

import click
from genologics.entities import Process

from cglims import api
from cglims.apptag import ApplicationTag

RELATION_UDS = ['motherID', 'fatherID', 'Other relations']
log = logging.getLogger(__name__)


@click.command()
@click.option('-s', '--source',
              type=click.Choice(['project', 'process']), default='project')
@click.argument('lims_id')
@click.pass_context
def samples(context, source, apptags, lims_id):
    """Fetch projects from the database."""
    lims = api.connect(context.obj)
    if source == 'process':
        lims_process = Process(lims, id=lims_id)
        lims_samples = process_samples(lims_process)
    elif source == 'project':
        lims_samples = ({'sample': sample} for sample in
                        lims.get_samples(projectlimsid=lims_id))
    for lims_sample in lims_samples:
        click.echo(lims_sample.id)


@click.command()
@click.option('-u', '--update', is_flag=True, help='update sample UDFs')
@click.option('-f', '--force', is_flag=True, help='update existing values')
@click.option('-v', '--version', type=int, help='application tag version')
@click.option('-s', '--source',
              type=click.Choice(['sample', 'project', 'process']),
              default='sample')
@click.argument('lims_id')
@click.pass_context
def check(context, update, version, force, source, lims_id):
    """Check LIMS sample or all samples in a process."""
    lims = api.connect(context.obj)
    if source == 'sample':
        lims_samples = [{'sample': lims.sample(lims_id)}]
    elif source == 'process':
        lims_process = Process(lims, id=lims_id)
        lims_samples = process_samples(lims_process)
    elif source == 'project':
        lims_samples = ({'sample': sample} for sample in
                        lims.get_samples(projectlimsid=lims_id))

    for sample in lims_samples:
        check_sample(lims, sample['sample'], lims_artifact=sample.get('artifact'),
                     update=update, version=version, force=force)


def check_sample(lims, lims_sample, lims_artifact=None, update=False, version=None,
                 force=False):
    """Check a LIMS sample and optionally update some UDFs."""
    results = []
    log.info("checking sample: %s (%s)", lims_sample.id, lims_sample.name)
    log.debug('checking sample name...')
    results.append(check_samplename(lims_sample))
    log.debug('checking duplicate external sample name...')
    results.append(check_duplicatename(lims, lims_sample))
    log.debug('checking capture kit (extenal sequencing)...')
    results.append(check_capturekit(lims_sample))
    log.debug('checking family members...')
    results.append(check_familymembers(lims, lims_sample))

    if update:
        log.debug('updating missing reads...')
        set_missingreads(lims_sample, force=force)
        log.debug('checking if update to trio tag is possible...')
        set_trioapptag(lims, lims_sample)
        if version:
            log.debug('updating application tag version...')
            set_apptagversion(lims_sample, version, force=force)

        if lims_artifact:
            if lims_artifact.qc_flag:
                log.warn("qc flag already set: %s", lims_artifact.qc_flag)

            if False in results:
                log.warn("sample check FAILED: %s", lims_sample.id)
                lims_artifact.qc_flag = 'FAILED'
            else:
                log.info("sample check PASSED: %s", lims_sample.id)
                lims_artifact.qc_flag = 'PASSED'
            lims_artifact.put()


def set_missingreads(lims_sample, force=False):
    """Set the 'Reads Missing (M)' UDF base on app tag."""
    raw_apptag = lims_sample.udf['Sequencing Analysis']
    app_tag = ApplicationTag(raw_apptag)
    target_amount = app_tag.reads
    missing_reads = lims_sample.udf.get('Reads missing (M)')
    if not force and missing_reads is not None:
        log.warn("missing reads already set: %s", missing_reads)
    else:
        lims_sample.udf['Reads missing (M)'] = target_amount
        log.info("updating reads missing")
        lims_sample.put()


def set_apptagversion(lims_sample, version, force=False):
    """Set the Application Tag Version for a sample."""
    current_version = lims_sample.udf.get('Application Tag Version')
    if not force and current_version:
        log.warn("application tag version already set: %s", current_version)
    else:
        lims_sample.udf['Application Tag Version'] = version
        log.info("updating application tag version: %s", version)
        lims_sample.put()


def set_trioapptag(lims, lims_sample):
    """Update the application tag if a WGS trio has been sent in."""
    related_samples = lims.case(lims_sample.udf['customer'],
                                lims_sample.udf['familyID'])
    if len(related_samples) == 3:
        log.debug("found three related samples")
        # Q: should more than 3 samples be allowed?
        app_tags = set(sample.udf['Sequencing Analysis'] for sample in
                       related_samples)
        allowed_tags = set(['WGSPCFC030', 'WGTPCFC030'])
        if len(app_tags.difference(allowed_tags)) == 0:
            # then we can update the application tag for the sample
            log.info("found 3 related samples with WGS application tag")
            log.info("updating to trio WGS tag: %s", lims_sample.id)
            lims_sample.udf['Sequencing Analysis'] == 'WGTPCFC030'
            lims_sample.put()


def check_samplename(lims_sample):
    """Check external sample name conventions.

    Sample name are allowed to contain alphanumeric characters
    AND '-' (dash). Nothing else.
    """
    no_dash = lims_sample.name.replace('-', '')
    if not no_dash.isalnum():
        log.error("sample name not OK: %s", lims_sample.name)
        return False
    else:
        return True


def check_duplicatename(lims, lims_sample):
    """Check if the same customer has sent in a sample with the same id."""
    result = True
    samples = lims.get_samples(name=lims_sample.name,
                               udf={'customer': lims_sample.udf['customer']})
    for other_sample in samples:
        if other_sample.id != lims_sample.id:
            # same sample id twice!
            if other_sample.udf.get('cancelled') == 'yes':
                log.info("duplicate sample but the other is cancelled")
            else:
                log.error("sample name duplicate: %s | %s", lims_sample.id,
                          other_sample.id)
                result = False
    return result


def check_capturekit(lims_sample):
    """Check if no capture kit doesn't exist for external samples."""
    app_tag = lims_sample.udf['Sequencing Analysis']
    if app_tag.startswith('EXX'):
        # now we need a capture kit to be filled in
        capture_kit = lims_sample.udf['Capture Library version']
        if capture_kit == 'NA':
            log.error("capture kit not filled in for external sample")
            return False
    return True


def check_familymembers(lims, lims_sample):
    """Check if sample has family memeber but no relations to them."""
    result = True
    related_samples = lims.case(lims_sample.udf['customer'],
                                lims_sample.udf['familyID'])
    if len(related_samples) > 1:
        # sample is part of a family/case
        samples = {}
        is_family = False
        for sample in related_samples:
            if any(sample.udf.get(udf_key) for udf_key in RELATION_UDS):
                is_family = True
            samples[sample.name] = sample

        if is_family:
            for sample in related_samples:
                for udf_key in RELATION_UDS:
                    related_id = sample.udf.get(udf_key)
                    if related_id:
                        if related_id not in samples:
                            log.error("related sample not part of family")
                            log.error("original sample, %s", lims_sample.id)
                            log.error("family: %s", lims_sample.udf['familyID'])
                            log.error("other sample (missing): %s", related_id)
                            result = False
                        else:
                            log.debug("found related sample: %s = %s", udf_key,
                                      samples[related_id].id)
        else:
            # perhaps a tumor/normal pair?
            tumor_samples = (sample.udf.get('tumor') == 'yes' for sample in
                             related_samples)
            if any(tumor_samples):
                log.debug("samples part of cancer combo")
            else:
                log.error("samples in 'family' not related, tumor/normal?")
                result = False
    return result


def process_samples(lims_process):
    """Retrieve LIMS input samples from a process."""
    for artifact in lims_process.all_inputs():
        for lims_sample in artifact.samples:
            yield {'sample': lims_sample, 'artifact': artifact}



# -*- coding: utf-8 -*-

from cglims.api import SamplesheetHandler

def test_samplesheet():
    samplesheethandler = SamplesheetHandler()

    assert samplesheethandler._get_placement_lane('4:1') == 4
#    assert samplesheethandler._get_index('G07 - D707-D507 (CTGAAGCT-CAGGACGT)') == 'CTGAAGCT-CAGGACGT'



# -*- coding: utf-8 -*-
from copy import deepcopy
import re

from dateutil.parser import parse as parse_date
from genologics.entities import Sample
from genologics.lims import Lims

from cglims.apptag import ApplicationTag
from cglims.constants import READS_PER_1X, SEX_MAP
from cglims.exc import MultipleSamplesError

SAMPLE_REF = 'hg19'

def connect(config):
    """Connect and return API reference."""
    api = ClinicalLims(config['host'], config['username'], config['password'])
    return api


class ClinicalSample(object):

    def __init__(self, lims_sample):
        """ Wrapper around the genologics Sample class

        Args:
            lims_sample (genologics.Sample): the sample instance to extend
        """
        self.lims = lims_sample
        self._apptag = ApplicationTag(self.lims.udf['Sequencing Analysis'])

    @property
    def apptag(self):
        """ Init an instance of ApplicationTag based on the sample's apptag

        Returns: ApplicationTag.
        """
        return self._apptag

    @property
    def pipeline(self):
        """ Determines in which pipeline the sample needs to be run.

        Returns (str): 'mip' or 'mwgs'
        """
        if self.lims.udf.get('tissue_type') != 'tumour':
            if self.apptag.is_human:
                return 'mip'
        if self.apptag.is_microbial:
            return 'mwgs'

        return None

    @property
    def sex(self):
        """Return human readable form of sex (Gender)."""
        return SEX_MAP.get(self.udf('Gender'), None)

    @property
    def ordered_reads(self):
        """Calculate ordered number of reads."""
        app_tag = self.udf('Sequencing Analysis')
        type_id = app_tag[-4]
        number = int(app_tag[-3:])
        if type_id == 'R':
            return number * 1000000
        elif type_id == 'K':
            return number * 1000
        elif type_id == 'C':
            # expect WGS
            return number * READS_PER_1X
        else:
            raise ValueError("unknown read type id: {}".format(type_id))

    @property
    def expected_reads(self):
        """Calculate expected number of reads to sequence."""
        return int(self.ordered_reads * .75)

    def udf(self, udf_key, default=None):
        """Get a sample UDF."""
        return self.lims.udf.get(udf_key, default)

    @property
    def sample_id(self):
        """Get the official sample id."""
        return self.udf('Clinical Genomics ID') or self.lims.id

    def to_dict(self, minimal=False):
        """Export data from the sample object."""
        if self.udf('customer') and self.udf('familyID'):
            case_id = '-'.join([self.udf('customer'), self.udf('familyID')])
        else:
            case_id = 'NA'

        data = deepcopy(self.lims.udf._lookup)
        data.update(dict(
            id=self.lims.id,
            # general sample id if imported from old TSL
            sample_id=self.sample_id,
            name=self.lims.name,
            project_name=self.lims.project.name,
            project_id=self.lims.project.id,
            case_id=case_id,
        ))

        if not minimal:
            data.update(dict(
                date_received=parse_date(self.lims.date_received),
                sex=self.sex,
                reads=self.ordered_reads,
                expected_reads=self.expected_reads,
                is_human=self.apptag.is_human,
                sequencing_type=self.apptag.sequencing_type,
                is_external=self.apptag.is_external,
                pipeline=self.pipeline or 'NA',
                is_production=(False if data['customer'] == 'cust000' else True),
                panels=(self.udf('Gene List').split(';') if self.udf('Gene List') else None),
            ))
        return data


class SamplesheetHandler(object):

    def _get_placement_lane(self, lane):
        """Parse out the lane information from an artifact.placement"""
        return int(lane.split(':')[0])

    def _get_index(self, label):
        """Parse out the sequence from a reagent label"""

        reagent_types = self.get_reagent_types(name=label)

        if len(reagent_types) > 1:
            raise ValueError("Expecting at most one reagent type. Got ({}).".format(len(reagent_types)))

        try:
            reagent_type = reagent_types.pop()
        except IndexError:
            return ''
        sequence = reagent_type.sequence

        match = re.match(r"^.+ \((.+)\)$", label)
        if match:
            assert match.group(1) == sequence

        return sequence

    def _get_reagent_label(self, artifact):
        """Get the first and only reagent label from an artifact"""
        labels = artifact.reagent_labels
        if len(labels) > 1:
            raise ValueError("Expecting at most one reagent label. Got ({}).".format(len(labels)))
        return labels[0] if labels else None

    def _get_non_pooled_artifacts(self, artifact):
        """Find the parent artifact of the sample. Should hold the reagent_label"""
        artifacts = []

        if len(artifact.samples) == 1:
            artifacts.append(artifact)
        else:
            for input in artifact.input_artifact_list():
                artifacts.extend(self._get_non_pooled_artifacts(input))

        return artifacts

    def samplesheet(self, flowcell):
        containers = self.get_containers(name=flowcell)

        if containers:
            container = containers[-1] # only take the last one. See ÖA#217.
            raw_lanes = sorted(container.placements.keys())
            for raw_lane in raw_lanes:
                lane = self._get_placement_lane(raw_lane)
                placement_artifact = container.placements[raw_lane]
                for artifact in self._get_non_pooled_artifacts(placement_artifact):
                    sample = artifact.samples[0] # we are assured it only has one sample
                    label = self._get_reagent_label(artifact)
                    index = self._get_index(label)
                    yield {
                        'fcid': flowcell,
                        'lane': lane,
                        'sample_id': sample.id,
                        'sample_ref': SAMPLE_REF,
                        'index': index,
                        'description': '',
                        'sample_name': sample.project.name,
                        'control': 'N',
                        'recipe': 'R1',
                        'operator': 'script',
                        'project': sample.project.name
                    }


class ClinicalLims(Lims, SamplesheetHandler):

    def case(self, customer, family_id):
        filters = {'customer': customer, 'familyID': family_id}
        samples = self.get_samples(udf=filters)
        return samples

    def sample(self, lims_id, is_cgid=False):
        """Get a unique sample from LIMS."""
        if is_cgid:
            udf_key = 'Clinical Genomics ID'
            lims_samples = self.get_samples(udf={udf_key: lims_id})
            if len(lims_samples) == 1:
                return lims_samples[0]
            elif len(lims_samples) > 1:
                matches_str = ', '.join(sample.id for sample in lims_samples)
                message = "'{}' matches: {}".format(lims_id, matches_str)
                raise MultipleSamplesError(message)
            else:
                # no matching samples
                return None
        else:
            lims_sample = Sample(self, id=lims_id)

        return lims_sample

    def is_delivered(self, lims_id):
        """Check if a sample has been delivered."""
        filters = dict(samplelimsid=lims_id, type="Analyte",
                       process_type="CG002 - Delivery")
        delivery_analytes = self.get_artifacts(**filters)
        if delivery_analytes:
            return delivery_analytes[0].parent_process.udf['Date delivered']
        else:
            return None

    def process_samples(lims_process):
        """Retrieve LIMS input samples from a process."""
        for artifact in lims_process.all_inputs():
            for lims_sample in artifact.samples:
                yield {'sample': lims_sample, 'artifact': artifact}

    def get_received_date(self, lims_id):
        lims_artifacts = self.get_artifacts(process_type='CG002 - Reception Control',
                                            samplelimsid=lims_id)
        for artifact in lims_artifacts:
            udf_key = 'date arrived at clinical genomics'
            if artifact.parent_process and artifact.parent_process.udf.get(udf_key):
                return artifact.parent_process.udf.get(udf_key)


def deliver(lims_sample):
    """Figure out how to deliver results for a sample.

    Answers:
        - whether to delivery to Scout
        - where to deliver raw data
        - wether we shoud delivery extra assets
    """
    extras = []
    # should the sample be uploaded in Scout?
    to_scout = 'scout' in lims_sample.udf.get('Data Analysis', 'notFound')
    # we should always deliver at least FASTQ files somewhere, but where?
    # the default is to deliver to Caesar
    payload = {'target': 'caesar'}
    customer = lims_sample.udf['customer']

    uppmax_project = lims_sample.udf.get('uppmax_project')
    if uppmax_project:
        payload = {'target': 'uppmax', 'project': uppmax_project}

    if customer == 'cust002':
        extras.append('bam')
    elif customer == 'cust009':
        # molecular health... "there's a script for that"
        payload = {}
    elif customer in ('cust008', 'cust016', 'cust019', 'cust020'):
        payload = {'target': 'sll-sthlm@medstore.sahlgrenska.gu.se'}
    elif customer == 'cust000':
        payload = {'target': 'rasta'}

    return {
        'scout': to_scout,
        'customer': customer,
        'raw_data': payload,
        'extras': extras
    }
